/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: table_def_pub.c,v 1.9 2010/01/24 10:58:32 jaffe Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/
/* $Id: table_def_pub.c,v 1.9 2010/01/24 10:58:32 jaffe Exp $ */
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*----------------------------------------------------------------------- */
/* Table definition class for FITS binary tables.                         */
/* This package creates and maintains table definition structures used to */
/* create FITS binary tables.                                             */
/* A new table type XXX requires: (mostly in table_*_def.c)               */
/*  1) a unique definition of XXX_TABLE in VLTItable.h                    */  
/*  2) prototype and definition of make_XXX_table() which includes XXX.def*/
/*     in VLTIprivate.h and table_def_priv.c                              */ 
/*  3) an XXX.def file defining the components of the table in ../include */
/*  4) add case for table type in make_table_def                          */
/*     in table_def_pub.c                                                 */
/*----------------------------------------------------------------------- */
#include "fitsio.h"
#include "VLTIprivate.h"


/* public functions */

table_def* make_table_def(int table_type, char  colsymb, int symcount)
/*----------------------------------------------------------------------- */
/*   Make a table_def object by type                                      */
/*   Table data storage is allocated                                      */
/*   Inputs:                                                              */
/*      table_type   Table type code as defined above                     */
/*      colsym       A character to be substituted in column names        */
/*                   0 => don't check                                     */
/*      symcount     The maximum integer to replace colsym with           */
/*   Returns:                                                             */
/*      A table_def* for the table definition structure;                  */
/*      NULL indicates failure                                            */
/*----------------------------------------------------------------------- */
{
  /* create by type */
  switch (table_type) {
  case SOURCE_TABLE:
    return make_SOURCE_table(colsymb, symcount);
    break;
  case  ARRAY_GEOMETRY_TABLE:
    return make_ARRAY_GEOMETRY_table(colsymb, symcount);
    break;
  case  ADAPTIVE_OPTICS_TABLE:
    return make_ADAPTIVE_OPTICS_table(colsymb, symcount);
    break;
  case  FLAG_TABLE:
    return make_FLAG_table(colsymb, symcount);
    break;
  case  DELAY_TABLE:
    return make_DELAY_table(colsymb, symcount);
    break;
  case  DISPERSION_TABLE:
    return make_DISPERSION_table(colsymb, symcount);
    break;
  case  IMAGING_DATA_TABLE:
    return make_IMAGING_DATA_table(colsymb, symcount);
    break;
  case  IMAGING_FDATA_TABLE:
    return make_IMAGING_FDATA_table(colsymb, symcount);
    break;
  case  IMAGING_DETECTOR_TABLE:
    return make_IMAGING_DETECTOR_table(colsymb, symcount);
    break;
  case  PREVISIBILITY_TABLE:
    return make_PREVISIBILITY_table(colsymb, symcount);
    break;
  case  VISIBILITY_TABLE:
    return make_VISIBILITY_table(colsymb, symcount);
    break;
  case  TIP_TILT_TABLE:
    return make_TIP_TILT_table(colsymb, symcount);
    break;
  case CALIBRATION_TABLE:
    return make_CALIBRATION_table(colsymb, symcount);
    break;
  case OPTICAL_TRAIN_TABLE:
    return make_OPTICAL_TRAIN_table(colsymb, symcount);
    break;
  case WAVELENGTH_TABLE:
    return make_WAVELENGTH_table(colsymb, symcount);
    break;
  case OI_VIS_TABLE:
    return make_OI_VIS_table(colsymb, symcount);
    break;
  case OI_TARGET_TABLE:
    return make_OI_TARGET_table(colsymb, symcount);
    break;
  case XXX_TABLE:   /* as an example only */
    return make_XXX_table(colsymb, symcount);
    break;
  default:  /* unknown type */
    fprintf(stderr,"ERROR: make_table_def: unknown table type code %d\n", 
	    table_type);
    return NULL;
  } /* END of switch by type */
} /* end  make_table_def */

void kill_table_def(table_def* tabledef)
/*----------------------------------------------------------------------- */
/*   Destroy a table_def object                                           */
/*   Inputs:                                                              */
/*      tabledef   Pointer to the structure to destroy.                   */
/*----------------------------------------------------------------------- */
{
  int i, nkey, ncol;

  if (tabledef==NULL) return; /* sanity check */

  nkey = tabledef->nkey;
  ncol = tabledef->ncol;

  /* free storage for parts */
  if (tabledef->extname!=NULL) free (tabledef->extname);
  
  /* free keywords */
 	if( tabledef->keywords)
 	{
 		for(i=0 ;i <nkey; i++)
 		{
 			if (tabledef->keywords[i] )
 				free (tabledef->keywords[i]);
 		}
 		free (tabledef->keywords);
 	}
 
 	if( tabledef->keycom)
 	{
 		for(i=0 ;i <nkey; i++)
 		{
 			if (tabledef->keycom[i] )
 				free (tabledef->keycom[i]);
 		}
 		free (tabledef->keycom);
 	}
 
 	if( tabledef->keydata)
 	{
 		for(i=0 ;i <nkey; i++)
 		{
 			if (tabledef->keydata[i] )
 				free (tabledef->keydata[i]);
 		}
 		free (tabledef->keydata);
 	}

  /* remove keywords structures */
  if (tabledef->keytype!=NULL)  free (tabledef->keytype);

  /* free columns */
 	if( tabledef->coltype)
 	{
 		for(i=0 ;i <ncol; i++)
 		{
 			if (tabledef->coltype[i] )
 				free (tabledef->coltype[i]);
 		}
 		free (tabledef->coltype);
 	}
 
 	if( tabledef->colunits)
 	{
 		for(i=0 ;i <ncol; i++)
 		{
 			if (tabledef->colunits[i] )
 				free (tabledef->colunits[i]);
 		}
 		free (tabledef->colunits);
 	}
 
 	if( tabledef->coldim)
 	{
 		for(i=0 ;i <ncol; i++)
 		{
 			if (tabledef->coldim[i] )
 				free (tabledef->coldim[i]);
 		}
 		free (tabledef->coldim);
 	}


  free (tabledef); /* finish it off */
} /* end  kill_table_def */

  void make_table_from_def (fitsfile *fptr, table_def *tabledef,
		      int *ierr)
/*----------------------------------------------------------------------- */
/*   Create and initialize a FITS binary table in an open FITS file       */
/*   Template values from the definitions are used which may be modified  */
/*   as necessary to get array dimensionality correct.                    */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      tabledef  definition structure for table                          */
/*   Outputs:                                                             */
/*      ierr      Error code, 0 = OK, else failed, checked on input       */
/*----------------------------------------------------------------------- */
{
  int i;
  char code;
  char** tform;
  char dimkey[10];

  if (*ierr>0) return; /* existing error condition? */

  /* create formats strings */
  tform = (char**)malloc (tabledef->ncol*sizeof(char**));
  for (i=0; i<tabledef->ncol; i++) {
    tform[i] = (char*)malloc(50);
    sprintf (tform[i],"%ld%c",
	     tabledef->colcount[i],
	     FITS_type_to_code(tabledef->coldatatype[i]));
  } /* end loop over formats */

  make_table (fptr, (long)0, tabledef->ncol, tabledef->coltype,
		   tform, tabledef->colunits, tabledef->extname,
		   tabledef->nkey, tabledef->keytype, tabledef->keywords, 
		   tabledef->keydata, tabledef->keycom, 
		   ierr);

  /* delete formats strings */
  for (i=0; i<tabledef->ncol; i++)  free (tform[i]);
  free (tform);
    
  if (*ierr!=0) return;  /* error check */

  /* add column dimensionalities */
  for (i=0; i<tabledef->ncol; i++) {
    if (tabledef->coldim[i]!=NULL) {
      if (tabledef->coldim[i][0]=='(') { /* simple validity check */
	sprintf (dimkey,"TDIM%d",i+1); /* keyword for column */
	fits_write_key(fptr, TSTRING, dimkey, (void*)tabledef->coldim[i],
		       "Column dimensionality ",
		       ierr);
      } /* end valid dimension */
    } /* end of not-null pointer */
  } /* end adding column dimensionalities */
} /* end make_table_from_def */

table_row* make_table_row_from_def (table_def *tabledef)
/*----------------------------------------------------------------------- */
/*   function to create a row structure from a table defintion.           */
/*   Inputs:                                                              */
/*      tabledef  definition structure for table                          */
/*   Returns:                                                             */
/*      Pointer to a fully created table_row structure,NULL=>failed       */
/*----------------------------------------------------------------------- */
{
  table_row* out;
  int ncol;
  long i, j, n, size;

  if (tabledef==NULL) return NULL; /* error check */
  ncol = tabledef->ncol;

  /* make basic framework */
  out = make_table_row(0, tabledef->ncol);
  if (out==NULL) return out; /* error check */

  /* finish column information */
  for (i=0; i<ncol; i++) { /* each col i */
    size = tabledef->colcount[i] * FITS_type_to_size(tabledef->coldatatype[i]);
    /* if string add byte for final NULL */
    if (tabledef->coldatatype[i]==TSTRING) size++; 
    out->data[i]->tdata.tvoid = (void*)malloc(size);   /* allocate column storage */
    /* initialize data */
    n = tabledef->colcount[i];
    switch (tabledef->coldatatype[i]) {
    case TSHORT:
      for (j=0;j<n;j++) out->data[i]->tdata.tshort[j] = 0;
      break;
    case TINT:
      for (j=0;j<n;j++) out->data[i]->tdata.tint[j] = 0;
      break;
    case TLONG:
      for (j=0;j<n;j++) out->data[i]->tdata.tlong[j] = 0;
      break;
    case TFLOAT:
      for (j=0;j<n;j++) out->data[i]->tdata.tfloat[j] = (float)0.0;
      break;
    case TCOMPLEX:
      for (j=0;j<n;j+=2) {
	out->data[i]->tdata.tfloat[j] = (float)0.0;
	out->data[i]->tdata.tfloat[j+1] = (float)0.0;
      }
      break;
    case TDOUBLE:
      for (j=0;j<n;j++) out->data[i]->tdata.tdouble[j] = 0.0;
      break;
    case TSTRING:
      for (j=0;j<n;j++) out->data[i]->tdata.tchar[j] = ' ';
      break;
    default:
      for (j=0;j<n;j++) out->data[i]->tdata.tchar[j] = 0;
    } /* end switch over data type */
  } /* end loop over columns */

  return out; /* done */
} /* end make_table_row_from_def */

table_def* make_table_def_from_table (fitsfile *fptr)
/*----------------------------------------------------------------------- */
/*   Make a table_def object from an open binary table                    */
/*   Inputs:                                                              */
/*      fptr         cfitsio pointer to open FITS file                    */
/*   Returns:                                                             */
/*      A table_def* for the table definition structure;                  */
/*      NULL indicates failure                                            */
/*----------------------------------------------------------------------- */
{
  table_def* out = NULL;
  char commnt[FLEN_COMMENT], keyword[FLEN_KEYWORD];
  char cdata[1][FLEN_CARD];  /* Strings different in cfitsio */
  int i, ierr=0;
  long ncol;
  char ttype[FLEN_VALUE], tunit[FLEN_VALUE], disp[FLEN_VALUE];
  char typechar[FLEN_VALUE];
  long repeat, nullval;
  double scale, zero;

  if (fptr==NULL) return out; /* sanity check */

  out = (table_def*)malloc(sizeof(table_def)); /* create structure */
  if (out==NULL) return out;  /* check allocation */
  /* initialize  */
  memset (out, 0, sizeof(table_def));

  /* binary table type */
  strncpy (cdata[0], "                ", 16);
  fits_read_key_str (fptr, "EXTNAME", cdata[0], commnt, &ierr);
  if (ierr!=0) {
    fprintf (stderr, "ERROR: make_table_def_from_table: Not a binary table.\n");
    fprintf (stderr, "error number %d.\n", ierr);
    return NULL; /* apparently not a bin table */
  } else {  /* save type */
    out->extname = (char*)malloc(strlen(cdata[0])+1);
    strcpy (out->extname, cdata[0]);
  }

  /* read any application specific keywords*/
  get_application_keywords(fptr, out, &ierr);
  if (ierr!=0) { /* problems reading/parsing table? */
    kill_table_def(out); /* destroy what you create */
    fprintf (stderr, "ERROR:  make_table_def_from_table reading keywords.\n");
    return NULL;
  }
  
  /* column info - cfitsio makes this easier */
  ncol = 0;
  fits_read_key_lng (fptr, "TFIELDS", &ncol, commnt, &ierr);
  if (ncol>0) {
    /* create arrays */
    out->coltype = (char**)malloc(ncol*sizeof(char*));
    out->colcount = (long*)malloc(ncol*sizeof(long));
    out->coldatatype = (int*)malloc(ncol*sizeof(int));
    out->colunits = (char**)malloc(ncol*sizeof(char*));
    out->coldim = (char**)malloc(ncol*sizeof(char*));
    /* fill arrays with column data */
    out->ncol = ncol;
    for (i=0; i<ncol; i++) { 
      /* get column info */
      fits_get_bcolparms (fptr, i+1, ttype, tunit, typechar,
			  &repeat, &scale, &zero, &nullval, disp, 
			  &ierr);
      out->coltype[i] = (char*)malloc(strlen(ttype)+1);
      strcpy(out->coltype[i], ttype);
      out->colcount[i] = repeat;
      out->coldatatype[i] = FITS_code_to_type(typechar[0]);
      out->colunits[i] = (char*)malloc(strlen(tunit)+1);
      strcpy(out->colunits[i], tunit);
      /* need to get TDIMn value */
      sprintf (keyword,"TDIM%d",i+1);
      strncpy (cdata[0], "                        ", 24); /* blank fill */
      fits_read_key_str (fptr, keyword, cdata[0], commnt, &ierr);
      ierr = 0;  /* ignore errors and use blanks */
      out->coldim[i] = (char*)malloc(strlen(cdata[0])+1);
      strcpy(out->coldim[i], cdata[0]);
    } /* end loop over columns */
  } /* end of positive column number section */
		      
  if (ierr!=0) { /* error reading column info? */
    fprintf (stderr, "make_table_def_from_table: Error %d getting column info",
	     ierr);
    return NULL; /* error */
  }
  
  return out; /* done */
} /* end make_table_def_from_table */

void table_get_keyword(table_def *tabledef, int keyword, void *value, 
		       int *ierr)
/*----------------------------------------------------------------------- */
/*   Function to Read a keyword value from a table definition             */
/*   Inputs:                                                              */
/*      tabledef   Pointer to the table definition structure              */
/*      keyword    1-rel keyword number                                   */
/*      value      pointer to memory to recieve value                     */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, returns if initially positive      */
/*                checked on input                                        */
/*----------------------------------------------------------------------- */
{
  if (*ierr>0) return; /* existing error condition? */
  if (tabledef==NULL) {*ierr=10000; return;} /* sanity check */
  if ((tabledef->nkey+1)<keyword) {*ierr=10000; return;} /* keyword in range*/
  if (keyword<1) {*ierr=10000; return;} /* keyword in range*/
  
  /* branch by keyword type, strings different */
  if (tabledef->keytype[keyword-1]==TSTRING) { /* string */
    memcpy (value, tabledef->keydata[keyword-1], 
	    strlen(tabledef->keydata[keyword-1])+1);
  } else { /* numeric */
    memcpy (value, tabledef->keydata[keyword-1],  
	    FITS_type_to_size(tabledef->keytype[keyword-1]));
  }
} /* end table_get_keyword*/

void table_put_keyword(table_def *tabledef, int keyword, void *value,
		       int *ierr)
/*----------------------------------------------------------------------- */
/*   Function to write a keyword value to a table definition              */
/*   Inputs:                                                              */
/*      tabledef   Pointer to the table definition structure              */
/*      keyword    1-rel keyword number                                   */
/*      value      pointer to memory containing value                     */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, returns if initially positive      */
/*----------------------------------------------------------------------- */
{
  if (*ierr>0) return; /* existing error condition? */
  if (tabledef==NULL) {*ierr=10000; return;} /* sanity check */
  if ((tabledef->nkey+1)<keyword) {*ierr=10000; return;} /* keyword in range*/
  if (keyword<1) {*ierr=10000; return;} /* keyword in range*/
  
  /* branch by keyword type, strings different */
  if (tabledef->keytype[keyword-1]==TSTRING) { /* string */
    memcpy (tabledef->keydata[keyword-1], value, 
	    strlen(value)+1);
  } else { /* numeric */
    memcpy (tabledef->keydata[keyword-1], value, 
	    FITS_type_to_size(tabledef->keytype[keyword-1]));
  }
} /* end table_put_keyword*/

/* */
